class PartialCommunity:
    """Represents a Weverse Community with partial information available.

    Inherited by:

    - :class:`Community`

    .. container:: operations

        .. describe:: x == y

            Checks if two communities are equal.

        .. describe:: x != y

            Checks if two communities are not equal.

        .. describe:: hash(x)

            Returns the community's hash.

        .. describe:: str(x)

            Returns the community's name.

    Attributes
    ----------
    data: :class:`dict`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`int`
        The ID of the community.
    name: :class:`str`
        The official name of the community.
    logo_image_url: :class:`str`
        The URL to the logo image of the community.
    """

    __slots__ = ("data", "id", "name", "logo_image_url")

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: int = data["communityId"]
        self.name: str = data["communityName"]
        self.logo_image_url: str = data["logoImage"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Partial Community community_id={self.id}, name={self.name}"

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.id)

    @property
    def url(self) -> str:
        """:class:`str`: Returns the URL of the Weverse Community's Artist tab."""
        return f"https://weverse.io/{self.data['urlPath']}/artist"


class Community(PartialCommunity):
    """Represents a Weverse Community. Inherits from :class:`PartialCommunity`.

    Shares the same attributes and operations with :class:`PartialCommunity`.

    Attributes
    ----------
    alias: :class:`str`
        The alias of the community.
    agency_profile_name: :class:`str`
        The profile name of the community's agency profile.
    agency_image_url: :class:`str`
        The URL of the profile image of the community's agency profile.
    agency_cover_image_url: :class:`str`
        The URL of the cover image of the community's agency profile.
    home_header_image_url: :class:`str`
        The URL of the home header of the community.
    home_gradation_colour: :class:`dict`
        The dictionary that contains the colour combinations displayed
        on the community's home page.
    has_membership: :class:`bool`
        Whether the community has paid membership benefits.
    member_count: :class:`int`
        The number of members who joined the community.
    fandom_name: :class:`str` | :class:`None`
        The fandom name of the community, if any.
    fan_event_url: :class:`str` | :class:`None`
        The URL used to participate in the fan events of the community, if any.
    artist_code: :class:`str` | :class:`None`
        The artist code of the community, if any.
    open_date: :class:`str` | :class:`None`
        The date the community got created on Weverse, if any.
    """

    __slots__ = (
        "alias",
        "agency_profile_name",
        "agency_image_url",
        "agency_cover_image_url",
        "home_header_image_url",
        "home_gradation_colour",
        "has_membership",
        "member_count",
        "fandom_name",
        "fan_event_url",
        "artist_code",
        "open_date",
    )

    def __init__(self, data: dict):
        super().__init__(data)
        self.alias: str = data["communityAlias"]
        self.agency_profile_name: str = data["agencyProfile"]["profileName"]
        self.agency_image_url: str = data["agencyProfile"]["profileImageUrl"]
        self.agency_cover_image_url: str = data["agencyProfile"]["profileCoverImageUrl"]
        self.home_header_image_url: str = data["homeHeaderImage"]
        self.home_gradation_colour: dict = data["homeGradationColor"]
        self.has_membership: bool = data["hasMembershipProduct"]
        self.member_count: int = data["memberCount"]
        self.fandom_name: str | None = data.get("fandomName")
        self.fan_event_url: str | None = data.get("fanEventUrl")
        self.artist_code: str | None = data.get("artistCode")
        self.open_date: str | None = data.get("openDate")

    def __repr__(self):
        return f"Community community_id={self.id}, name={self.name}"



from .media import WeverseMedia


class Live(WeverseMedia):
    """Represents a Weverse Live Broadcast.

    Inherits from :class:`.media.WeverseMedia`.

    Shares the same attributes with :class:`.post.PostLike`,
    :class:`.media.MediaLike` and :class:`.media.WeverseMedia`.

    Attributes
    ----------
    message_count: :class:`int` | :class:`None`
        The number of messages in the live broadcast, if any.
    """

    __slots__ = ("message_count",)

    def __init__(self, data: dict):
        super().__init__(data)
        self.message_count: int | None = (
            data["extension"]["mediaInfo"]["chat"]["messageCount"]
            if data["extension"]["mediaInfo"].get("chat")
            else None
        )

    def __repr__(self):
        return f"Live live_id={self.id}, title={self.title}"



class WeverseException(Exception):
    """Base exception class that other Weverse exceptions inherit from."""


class LoginError(WeverseException):
    """An Exception raised when the login to Weverse fails."""

    def __init__(self, code: int, reason: str):
        message = f"Login to Weverse has failed.\nCode: {code}\nReason: {reason}"
        super().__init__(message)


class RequestFailed(WeverseException):
    """An Exception raised when the status code of the response returned by the
    API is not 200, 401, 403, 404 or 500. (Note: This is because 401, 403, 404
    and 500 has their own specific exceptions.)"""

    def __init__(self, url: str, code: int, reason: str):
        message = f"Request to {url} has failed.\nCode: {code}\nReason: {reason}."
        super().__init__(message)


class TokenExpired(WeverseException):
    """An Exception raised when the status code of the response returned by the
    API is 401."""

    def __init__(self, url: str):
        message = (
            f"Request to {url} has failed.\nCode: 401\n\n"
            "Reason: The access token has potentially expired."
        )
        super().__init__(message)


class Forbidden(WeverseException):
    """An Exception raised when the status code of the response returned by the
    API is 403."""

    def __init__(self, url: str, reason: str):
        message = f"Request to {url} has failed.\nCode: 403\nReason: {reason}."
        super().__init__(message)


class NotFound(WeverseException):
    """An Exception raised when the status code of the response returned by the
    API is 404."""

    def __init__(self, url: str):
        message = (
            f"Request to {url} has failed.\nCode: 404\n"
            "Reason: The requested resource cannot be found or does not exist."
        )
        super().__init__(message)


class InternalServerError(WeverseException):
    """An Exception raised when the status code of the response returned by the
    API is 500."""

    def __init__(self, url: str):
        message = (
            f"Request to {url} has failed.\nCode: 500\n\n"
            "Reason: The Weverse API encountered an Internal Server Error."
        )
        super().__init__(message)



import asyncio
import logging

from aiohttp import ClientConnectionError, ClientResponse, ClientSession
from yarl import URL

import weverse.errors
import weverse.url

from .utils import MISSING

logger = logging.getLogger("__main__")


class WeverseFetcher:
    """The Weverse Fetcher that fetches the raw data from Weverse's API.

    Attributes
    ----------
    login_payload: :class:`dict`
        The dictionary that contains the login payload to be used when
        retrieving the access token.
    """

    def __init__(self, login_payload: dict):
        self.__login_payload: dict = login_payload
        self.__access_token: str = MISSING

    @property
    def api_header(self) -> dict:
        """:class:`dict`: Returns the API headers used to communicate with Weverse."""
        return {
            "referer": "https://weverse.io",
            "authorization": f"Bearer {self.__access_token}",
        }

    async def fetch_access_token(self) -> None:
        """Retrieves the access token to be used to interact with Weverse's API.

        Raises
        ------
        LoginError
            The login credentials are invalid.
        """
        login_headers = {
            "content-type": "application/json",
            "x-acc-app-secret": "5419526f1c624b38b10787e5c10b2a7a",
            "x-acc-app-version": "abc",
            "x-acc-language": "en",
            "x-acc-service-id": "weverse",
            "x-acc-trace-id": "abc",
        }
        login_url = "https://accountapi.weverse.io/web/api/v2/auth/token/by-credentials"

        async with ClientSession() as client, client.post(
            login_url, headers=login_headers, json=self.__login_payload
        ) as resp:
            if resp.status != 200:
                raise weverse.errors.LoginError(
                    resp.status, (await resp.json()).get("message")
                )

            data = await resp.json()
            self.__access_token = data["accessToken"]
            logger.debug(
                "LOGIN SUCCESSFUL\nData: %s\nAccess Token: %s",
                data,
                self.__access_token,
            )

    async def __status_code_checker(self, response: ClientResponse) -> None:
        """Checks the response code and throws the relevant exceptions
        accordingly if the response code is not 200.

        Parameters
        ----------
        response: :class:`aiohttp.ClientResponse`
            The response to check for.

        Raises
        ------
        TokenExpired
            If the status code is 401.
        Forbidden
            If the status code is 403.
        NotFound
            If the status code is 404.
        InternalServerError
            If the status code is 500.
        RequestFailed
            If the status code is anything that is not 200, 401, 404 and 500.
        """
        if response.status == 200:
            return

        if response.status == 401:
            raise weverse.errors.TokenExpired(response.url)

        if response.status == 403:
            raise weverse.errors.Forbidden(
                response.url, (await response.json()).get("message")
            )

        if response.status == 404:
            raise weverse.errors.NotFound(response.url)

        if response.status == 500:
            raise weverse.errors.InternalServerError(response.url)

        raise weverse.errors.RequestFailed(
            response.url, response.status, (await response.json()).get("message")
        )

    async def __fetch(self, url: str) -> dict:
        """The generic fetcher that fetches data from Weverse's API.

        Parameters
        ----------
        url: :class:`str`
            The URL that is used to fetch data from.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the raw data fetched from Weverse's API.
        """
        error_count = 0
        max_error_count = 3

        while error_count < max_error_count:
            try:
                async with ClientSession() as client, client.get(
                    url=URL(url, encoded=True), headers=self.api_header
                ) as resp:
                    await self.__status_code_checker(resp)
                    data = await resp.json()
                    logger.debug(
                        "RESOURCE SUCCESSFULLY FETCHED FROM WEVERSE.\nDATA: %s", data
                    )
                    return data

            except weverse.errors.TokenExpired:
                error_count += 1
                logger.warning(
                    "ACCESS TOKEN POTENTIALLY EXPIRED. "
                    "ATTEMPTING TO RENEW ACCESS TOKEN AND RETURN "
                    "REQUESTED RESOURCE SILENTLY.\n"
                    "Affected URL: %s\nError Count: %s/%s",
                    url,
                    error_count,
                    max_error_count,
                )
                await self.fetch_access_token()

            except (asyncio.exceptions.TimeoutError, ClientConnectionError):
                error_count += 1
                logger.warning(
                    "CLIENT CONNECTION ERROR ENCOUNTERED. "
                    "ATTEMPTING TO RETURN REQUESTED RESOURCE SILENTLY.\n"
                    "Affected URL: %s\nError Count: %s/%s",
                    url,
                    error_count,
                    max_error_count,
                )

            except weverse.errors.InternalServerError:
                error_count += 1
                logger.warning(
                    "TEMPORARY INTERNAL SERVER ERROR ENCOUNTERED. "
                    "ATTEMPTING TO RETURN REQUESTED RESOURCE SILENTLY.\n"
                    "Affected URL: %s\nError Count: %s/%s",
                    url,
                    error_count,
                    max_error_count,
                )

            await asyncio.sleep(1)

    async def fetch_latest_notifications(self) -> dict:
        """Retrieves the dictionary that contains the data of the latest
        Weverse notifications from Weverse's API.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the latest Weverse
            notifications.
        """
        return await self.__fetch(weverse.url.latest_notifications_url())

    async def fetch_notification(self, notification_id: int) -> dict:
        """Retrieves the dictionary that contains the data of the specified
        Weverse Notification from Weverse's API.

        Parameters
        ----------
        notification_id: :class:`int`
            The notification ID of the notification to fetch.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the specified Weverse notification.

        Raises
        ------
        NotFound
            If the requested Weverse Notification does not exist.
        """
        url = weverse.url.notification_url(notification_id)
        data = await self.__fetch(url)

        if not data["data"] or not data["data"][0].get("community"):
            raise weverse.errors.NotFound(url)

        return data["data"][0]

    async def fetch_joined_communities(self) -> dict:
        """Retrieves the dictionary that contains the data of the joined
        communities the account has access to, from Weverse's API.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the joined communities.
        """
        return await self.__fetch(weverse.url.joined_communities_url())

    async def fetch_community(self, community_id: int) -> dict:
        """Retrieves the dictionary that contains the data of the specified
        Weverse Community from Weverse's API.

        Parameters
        ----------
        community_id: :class:`int`
            The community ID of the community to fetch.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the specified Weverse community.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community.
        NotFound
            If the requested Weverse Community does not exist.
        """
        return await self.__fetch(weverse.url.community_url(community_id))

    async def fetch_artists(self, community_id: int) -> dict:
        """Retrieves the dictionary that contains the data of the artists
        that belong to the specified Weverse Community from Weverse's API.

        Parameters
        ----------
        community_id: :class:`int`
            The community ID of the community to fetch the artists' data from.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the artists of the specified
            Weverse community.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community.
        NotFound
            If the requested Weverse Community does not exist.
        """
        return await self.__fetch(weverse.url.artists_url(community_id))

    async def fetch_post(self, post_id: str) -> dict:
        """Retrieves the dictionary that contains the data of the specified
        Weverse Post from Weverse's API.

        Parameters
        ----------
        post_id: :class:`str`
            The post ID of the post to fetch.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the specified Weverse post.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the post belongs to.
        NotFound
            If the requested Weverse Post does not exist.
        """
        return await self.__fetch(weverse.url.post_url(post_id))

    async def fetch_video_url(self, video_id: str) -> dict:
        """Fetches a Weverse Post Video URL by its ID.

        Parameters
        ----------
        video_id: :class:`str`
            The ID of the video to fetch.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the specified Weverse video.
        """
        return await self.__fetch(weverse.url.video_url(video_id))

    async def fetch_notice(self, notice_id: int) -> dict:
        """Retrieves the dictionary that contains the data of the specified
        Weverse Notice from Weverse's API.

        Parameters
        ----------
        notice_id: :class:`int`
            The notice ID of the notice to fetch.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the specified Weverse notice.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the notice belongs to.
        NotFound
            If the requested Weverse Notice does not exist.
        """
        return await self.__fetch(weverse.url.notice_url(notice_id))

    async def fetch_member(self, member_id: str) -> dict:
        """Retrieves the dictionary that contains the data of the specified
        Weverse Member from Weverse's API.

        Parameters
        ----------
        member_id: :class:`str`
            The member ID of the member to fetch.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the specified Weverse member.

        Raises
        ------
        NotFound
            If the requested Weverse Member does not exist.
        """
        return await self.__fetch(weverse.url.member_url(member_id))

    async def fetch_comment(self, comment_id: str) -> dict:
        """Retrieves the dictionary that contains the data of the specified
        Weverse Comment from Weverse's API.

        Parameters
        ----------
        comment_id: :class:`str`
            The comment ID of the comment to fetch.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the specified Weverse comment.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the comment belongs to.
        NotFound
            If the requested Weverse Comment does not exist.
        """
        return await self.__fetch(weverse.url.comment_url(comment_id))

    async def fetch_artist_comments(self, post_id: str) -> dict:
        """Retrieves the dictionary that contains the data of the artist comments
        that belong to the specified Weverse Post from Weverse's API.

        Parameters
        ----------
        post_id: :class:`str`
            The post ID of the post to fetch the artist comments' data from.

        Returns
        -------
        :class:`dict`
            The dictionary that contains the data of the artist comments of the
            specified Weverse post.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the post belongs to.
        NotFound
            If the requested Weverse Post does not exist.
        """
        return await self.__fetch(weverse.url.artist_comments_url(post_id))



from enum import Enum


class NotificationType(Enum):
    """The enum for Weverse Notification Types."""
    POST = "post"
    ARTIST_POST_COMMENT = "artist_post_comment"
    USER_POST_COMMENT = "user_post_comment"
    MEDIA_COMMENT = "media_comment"
    MOMENT_COMMENT = "moment_comment"
    MOMENT = "moment"
    LIVE = "live"
    NOTICE = "notice"
    MEDIA = "media"
    BIRTHDAY = "birthday"

    def __repr__(self):
        return self.value



import base64
import hmac
import time
import urllib.parse
from hashlib import sha1

WEVERSE_KEY = "1b9cb6378d959b45714bec49971ade22e6e24e42"
BASE_API_URL = "https://global.apis.naver.com/weverse/wevweb"
API_PARAMETERS = (
    "?appId=be4d79eb8fc7bd008ee82c8ec4ff6fd4&language=en&platform=WEB&wpf=pc"
)


def get_current_epoch_time() -> int:
    """Returns the current epoch time in milli seconds."""
    return int(time.time() * 1000)


def generate_message_digest(message: bytes) -> str:
    """Generates the message digest for Weverse.

    Parameters
    ----------
    message: :class:`bytes`
        The message to generate the digest for, in bytes.

    Returns
    -------
    :class:`str`
        The message digest made URL-compatible.
    """
    key = bytes(WEVERSE_KEY, "utf-8")
    hashed = hmac.new(key=key, msg=message, digestmod=sha1)
    message_digest = base64.b64encode(hashed.digest())
    return urllib.parse.quote_plus(message_digest)


def form_request_url(url: str) -> str:
    """Forms the final URL to be used to fetch data from Weverse's API.

    Parameters
    ----------
    url: :class:`str`
        The part of the URL that is after the base API URL.

    Returns
    -------
    :class:`str`
        The final URL that can be used to interact with Weverse's API.
    """
    indexed_url = url[0:255]
    epoch_time = get_current_epoch_time()

    url_with_epoch = f"{indexed_url}{epoch_time}".encode("utf-8")

    message_digest = generate_message_digest(url_with_epoch)
    return f"{BASE_API_URL}{url}&wmsgpad={epoch_time}&wmd={message_digest}"


def latest_notifications_url() -> str:
    """The endpoint for fetching the data of the latest notifications
    from Weverse."""
    url = "/noti/feed/v1.0/activities" + API_PARAMETERS
    return form_request_url(url)


def joined_communities_url() -> str:
    """The endpoint for fetching the data from Weverse
    that contains the communities the user have joined."""
    url = "/noti/feed/v1.0/activities/community" + API_PARAMETERS
    return form_request_url(url)


def community_url(community_id: int) -> str:
    """The endpoint for fetching the data of a specified Community from Weverse."""
    url = (
        f"/community/v1.0/community-{community_id}"
        + API_PARAMETERS
        + "&fieldSet=communityHomeV1"
    )
    return form_request_url(url)


def artists_url(community_id: int) -> str:
    """The endpoint for fetching the data of the artists of a specified Community
    from Weverse."""
    url = (
        f"/member/v1.0/community-{community_id}/artistMembers"
        + API_PARAMETERS
        + "&fieldSet=artistMembersV1&fields=communityId"
        "%2CjoinedDate%2CprofileType%2CprofileName%2CprofileImageUrl"
        "%2CprofileCoverImageUrl%2CprofileComment"
    )
    return form_request_url(url)


def post_url(post_id: str) -> str:
    """The endpoint for fetching the data of a specified Post from Weverse."""
    url = f"/post/v1.0/post-{post_id}" + API_PARAMETERS + "&fieldSet=postV1"
    return form_request_url(url)


def video_url(video_id: str) -> str:
    """The endpoint for fetching the URL of videos from Weverse."""
    url = f"/cvideo/v1.0/cvideo-{video_id}/downloadInfo" + API_PARAMETERS
    return form_request_url(url)


def notice_url(notice_id: int) -> str:
    """The endpoint for fetching the data of a specified Notice from Weverse."""
    url = f"/notice/v1.0/notice-{notice_id}" + API_PARAMETERS + "&fieldSet=noticeV1"
    return form_request_url(url)


def member_url(member_id: str) -> str:
    """The endpoint for fetching the data of a specified Member from Weverse."""
    url = (
        f"/member/v1.0/member-{member_id}"
        + API_PARAMETERS
        + "&fields=memberId%2CcommunityId%2Cjoined%2CjoinedDate%2CprofileType"
        "%2CprofileName%2CprofileImageUrl%2CprofileCoverImageUrl%2CprofileComment"
        "%2Chidden%2Cblinded%2CmemberJoinStatus%2CfollowCount%2ChasMembership"
        "%2ChasOfficialMark%2CfirstJoinAt%2Cfollowed%2CartistOfficialProfile%2CmyProfile"
    )
    return form_request_url(url)


def notification_url(notification_id: int) -> str:
    """The endpoint for fetching the data of a specified Notification
    from Weverse."""
    url = "/noti/feed/v1.0/activities" + API_PARAMETERS + f"&next={notification_id+1}"
    return form_request_url(url)


def comment_url(comment_id: str) -> str:
    """The endpoint for fetching the data of a specified Comment from Weverse."""
    url = f"/comment/v1.0/comment-{comment_id}" + API_PARAMETERS + "&fieldSet=commentV1"
    return form_request_url(url)


def artist_comments_url(post_id: str) -> str:
    """The endpoint for fetching the data of the artist comments on a
    specified post from Weverse."""
    url = (
        f"/comment/v1.0/post-{post_id}/artistComments"
        + API_PARAMETERS
        + "&fieldSet=postArtistCommentsV1"
    )
    return form_request_url(url)


# No appropriate use for it as of now.
# def latest_fan_posts_url(community_id: int) -> str:
#    """The endpoint for fetching the data of the latest fan posts
#    from Weverse."""
#    url = (
#        f"/post/v1.0/community-{community_id}/feedTab"
#        + API_PARAMETERS
#        + "&fields=feedTabPosts.fieldSet%28postsV1%29.limit%2820%29%2CcontentSlots."
#        "fieldSet%28contentSlotV1%29&platform=WEB"
#    )
#    return form_request_url(url)



import asyncio
import logging

from .enums import NotificationType
from .errors import NotFound
from .fetcher import WeverseFetcher
from .objects.comment import Comment
from .objects.community import Community, PartialCommunity
from .objects.live import Live
from .objects.media import ImageMedia, WeverseMedia, YoutubeMedia
from .objects.member import Artist, Member
from .objects.moment import Moment, OldMoment
from .objects.notice import Notice
from .objects.notification import Notification
from .objects.post import Post
from .url import get_current_epoch_time
from .utils import MISSING

logger = logging.getLogger(__name__)


class WeverseClient:
    """The Weverse Client that interacts with the official Weverse API.

    Attributes
    ----------
    email: :class:`str`
        The email of the Weverse account to login with.
    password: :class:`str`
        The password of the Weverse account to login with.
    stream_interval: :class:`float`
        The interval in querying the Weverse API for new notifications, in seconds.
    """

    def __init__(
        self,
        email: str,
        password: str,
        stream_interval: float = 20,
    ):
        self.email: str = email
        self.password: str = password
        self.stream_interval: float = stream_interval

        self.__fetcher: WeverseFetcher = WeverseFetcher(self.login_payload)
        self.__cache: _WeverseCache = _WeverseCache()
        self.__task: asyncio.Task = MISSING

    @property
    def login_payload(self) -> dict:
        """:class:`dict`: Returns the login payload used to login to Weverse."""
        return {"email": self.email, "password": self.password}

    async def fetch_latest_notifications(self) -> list[Notification]:
        """Fetches the latest notifications from Weverse's API.

        Returns
        -------
        list[:class:`.notification.Notification`]
            A list that contains the :class:`.notification.Notification` objects
            of the latest Weverse notifications.
        """
        data = await self.__fetcher.fetch_latest_notifications()
        data = data["data"]
        notifications = [
            Notification(notification_data)
            for notification_data in data
            if "community" in notification_data
        ]
        return notifications

    async def fetch_notification(self, notification_id: int) -> Notification:
        """Fetches a Weverse Notification from Weverse's API using its ID.

        Parameters
        ----------
        notification_id: :class:`int`
            The notification ID of the notification to fetch.

        Returns
        -------
        :class:`.notification.Notification`
            The :class:`.notification.Notification` object of the requested
            Weverse Notification.

        Raises
        ------
        NotFound
            If the requested Weverse Notification does not exist.
        """
        data = await self.__fetcher.fetch_notification(notification_id)
        return Notification(data)

    async def fetch_new_notifications(self) -> tuple[list[Notification], list[Comment]]:
        """Fetches the list of new Weverse Notifications from all
        communities that the signed-in account has access to.

        Returns
        -------
        tuple[list[:class:`.notification.Notification`], list[:class:`.comment.Comment`]]
            A tuple that contains a list of new :class:`.notification.Notification`
            objects and a list of new :class:`.comment.Comment` objects.

        Notes
        -----
        The reason why :class:`.comment.Comment` objects are returned for comment-related
        notifications and not :class:`.notification.Notification` objects is because Weverse
        does not provide an easy way to determine which comment-related
        notifications are new. As such, in the post-processing of
        comment-related notifications, the actual :class:`.comment.Comment`
        objects have to be fetched so that we know which comments are new.
        """
        old_non_comment_noti = dict(self.__cache.notification_cache)
        old_comment_noti = dict(self.__cache.comment_notification_cache)

        latest_notifications = await self.fetch_latest_notifications()
        (
            latest_non_comment_noti,
            latest_comment_noti,
        ) = self.__sort_notifications(latest_notifications)

        new_non_comment_noti = self._get_new_non_comment_noti(
            old_non_comment_noti, latest_non_comment_noti
        )

        new_artist_comments = await self._get_new_comment(
            old_comment_noti, latest_comment_noti
        )

        return new_non_comment_noti, new_artist_comments

    def __sort_notifications(
        self, notifications: list[Notification]
    ) -> tuple[list[Notification], list[Notification]]:
        """Sorts a dictionary containing notification data into
        comment notifications and non-comment notifications.

        Parameters
        ----------
        notifications: list[:class:`Notification`]
            The list that contains the :class:`Notification` objects of the
            notifications to sort.

        Returns
        -------
        tuple[list[:class:`Notification`], list[:class:`Notification`]]
            A tuple that contains a list of non-comment notifications and
            a list of comment notifications.
        """
        comment_notification_types = [
            NotificationType.USER_POST_COMMENT,
            NotificationType.MEDIA_COMMENT,
            NotificationType.MOMENT_COMMENT,
            NotificationType.ARTIST_POST_COMMENT,
        ]
        non_comment_notifications = []
        comment_notifications = []

        for notification in notifications:
            if notification.post_type in comment_notification_types:
                self.__cache.cache_comment_noti(notification)
                comment_notifications.append(notification)

            else:
                self.__cache.cache_notification(notification)
                non_comment_notifications.append(notification)

        return non_comment_notifications, comment_notifications

    @staticmethod
    def _get_new_non_comment_noti(
        old_cache: dict[int, Notification], notifications: list[Notification]
    ) -> list[Notification]:
        """Returns a list of new non-comment notifications that do not exist in
        the cache.

        Parameters
        ----------
        old_cache: dict[:class:`int`, :class:`Notificiation`]
            The old notification cache used to check which notifications are new.
        notifications: list[:class:`Notification`]
            The list that contains the latest notifications the user has access to.

        Returns
        -------
        list[:class:`Notification`]
            A list of new notifications.
        """
        current_epoch_time = get_current_epoch_time()
        epoch_time_difference = 600000  # Equivalent to 10 minutes.
        return [
            notification
            for notification in notifications
            if notification.id not in old_cache
            and current_epoch_time - notification.time_created <= epoch_time_difference
        ]

    async def _get_new_comment(
        self, old_cache: dict[str, int], notifications: list[Notification]
    ) -> list[Comment]:
        """Returns a list of new artist comments.

        Parameters
        ----------
        old_cache: dict[:class:`str`, :class:`int`]
            The old notification cache used to check which notifications are new.
        notifications: list[:class:`Notification`]
            The list that contains the latest notifications the user has access to.

        Returns
        -------
        list[:class:`Comment`]
            A list of :class:`Comment` objects of the new
            artist comments.
        """
        new_artist_comments = []
        current_epoch_time = get_current_epoch_time()
        epoch_time_difference = 60000  # Equivalent to 1 minute

        for notification in notifications:
            await asyncio.sleep(0.35)

            try:
                artist_comments = await self.fetch_artist_comments(notification.post_id)

            except NotFound:  # This is because the Post has been deleted.
                continue

            if not old_cache.get(notification.post_id + notification.author.id):
                for artist_comment in artist_comments:
                    if (
                        current_epoch_time - artist_comment.created_at
                        <= epoch_time_difference
                    ) and artist_comment.id not in self.__cache.comment_cache:
                        new_artist_comments.insert(0, artist_comment)
                        self.__cache.cache_comment(artist_comment)

            else:
                # This is done because there can be different artists that
                # have written on a post. As such, a new list is created
                # to contain only comments written by the author provided in
                # the notification data.
                artist_comments = [
                    artist_comment
                    for artist_comment in artist_comments
                    if artist_comment.author.id == notification.author.id
                ]
                comment_count = old_cache[notification.post_id + notification.author.id]
                new_comment_count = notification.count - comment_count

                for artist_comment in artist_comments[0:new_comment_count]:
                    if artist_comment.id not in self.__cache.comment_cache:
                        new_artist_comments.insert(0, artist_comment)
                        self.__cache.cache_comment(artist_comment)

        return new_artist_comments

    async def fetch_joined_communities(self) -> list[PartialCommunity]:
        """Fetches the joined communities from Weverse's API.

        Returns
        -------
        list[:class:`.community.PartialCommunity`]
            A list that contains the :class:`.community.PartialCommunity` objects
            of the joined Weverse communities.

        Notes
        -----
        If more information is required about the communities, you can
        make use of the :meth:`fetch_community()` method.
        """
        data = await self.__fetcher.fetch_joined_communities()
        data = data["data"]
        communities = [PartialCommunity(community_data) for community_data in data]
        return communities

    async def fetch_community(self, community_id: int) -> Community:
        """Fetches a Weverse Community from Weverse's API using its ID.

        Parameters
        ----------
        community_id: :class:`int`
            The community ID of the community to fetch.

        Returns
        -------
        :class:`.community.Community`
            The :class:`.community.Community` object of the requested Weverse
            Community.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community.
        NotFound
            If the requested Weverse Community does not exist.
        """
        data = await self.__fetcher.fetch_community(community_id)
        return Community(data)

    async def fetch_artists(self, community_id: int) -> list[Artist]:
        """Fetches a list of artists that belong to a Weverse Community
        from Weverse's API using the community ID.

        Parameters
        ----------
        community_id: :class:`int`
            The community ID of the community to fetch the artists from.

        Returns
        -------
        list[:class:`.member.Artist`]
            A list that contains the :class:`.member.Artist` objects of the specified
            Weverse community.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community.
        NotFound
            If the requested Weverse Community does not exist.
        """
        data = await self.__fetcher.fetch_artists(community_id)
        artists = [Artist(artist_data) for artist_data in data]
        return artists

    async def fetch_post(self, post_id: str) -> Post:
        """Fetches a Weverse Post from Weverse's API using its ID.

        Parameters
        ----------
        post_id: :class:`str`
            The post ID of the post to fetch.

        Returns
        -------
        :class:`.post.Post`
            The :class:`.post.Post` object of the requested Weverse Post.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the post belongs to.
        NotFound
            If the requested Weverse Post does not exist.
        """
        data = await self.__fetcher.fetch_post(post_id)
        return Post(data)

    async def fetch_media(
        self, media_id: str
    ) -> ImageMedia | WeverseMedia | YoutubeMedia:
        """Fetches a Weverse Media from Weverse's API using its ID.

        Parameters
        ----------
        media_id: :class:`str`
            The media ID of the media to fetch.

        Returns
        -------
        :class:`.media.ImageMedia` | :class:`.media.WeverseMedia` | :class:`.media.YoutubeMedia`
            Can be either a :class:`.media.ImageMedia`, :class:`.media.WeverseMedia` or
            :class:`.media.YoutubeMedia` object of the requested Weverse Media depending
            on the type of media it is.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the media belongs to.
        NotFound
            If the requested Weverse Media does not exist.
        """
        data = await self.__fetcher.fetch_post(media_id)

        if "image" in data["extension"]:
            return ImageMedia(data)

        if "video" in data["extension"]:
            return WeverseMedia(data)

        return YoutubeMedia(data)

    async def fetch_live(self, live_id: str) -> Live:
        """Fetches a Weverse Live Broadcast from Weverse's API using its ID.

        Parameters
        ----------
        live_id: :class:`str`
            The live ID of the live broadcast to fetch.

        Returns
        -------
        :class:`.live.Live`
            The :class:`.live.Live` object of the requested Weverse Live Broadcast

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the live broadcast belongs to.
        NotFound
            If the requested Weverse Live Broadcast does not exist.
        """
        data = await self.__fetcher.fetch_post(live_id)
        return Live(data)

    async def fetch_moment(self, moment_id: str) -> Moment | OldMoment:
        """Fetches a Weverse Moment from Weverse's API using its ID.

        Parameters
        ----------
        moment_id: :class:`str`
            The moment ID of the moment to fetch.

        Returns
        -------
        :class:`.moment.Moment` | :class:`.moment.OldMoment`
            The :class:`.moment.Moment` or :class:`.moment.OldMoment` object
            of the requested Weverse moment depending on the type of moment it is.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the moment belongs to.
        NotFound
            If the requested Weverse Moment does not exist.
        """
        data = await self.__fetcher.fetch_post(moment_id)

        if "moment" in data["extension"]:
            return Moment(data)

        return OldMoment(data)

    async def fetch_video_url(self, video_id: str) -> str:
        """Fetches a Weverse Post Video URL by its ID.

        Parameters
        ----------
        video_id: :class:`str`
            The ID of the video to fetch.

        Returns
        -------
        :class:`str`
            The URL of the video.

        Notes
        -----
        An additional API call has to be made because unlike images,
        Weverse does include the URL to videos in the response for posts.
        """
        data = await self.__fetcher.fetch_video_url(video_id)
        url_mapping = {}

        for video_data in data["downloadInfo"]:
            url_mapping[int(video_data["resolution"].replace("P", ""))] = video_data[
                "url"
            ]

        return url_mapping[max(url_mapping)]

    async def fetch_notice(self, notice_id: int) -> Notice:
        """Fetches a Weverse Notice from Weverse's API using its ID.

        Parameters
        ----------
        notice_id: :class:`int`
            The notice ID of the notice to fetch.

        Returns
        -------
        :class:`.notice.Notice`
            The :class:`.notice.Notice` object of the requested Weverse Notice.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the notice belongs to.
        NotFound
            If the requested Weverse Notice does not exist.
        """
        data = await self.__fetcher.fetch_notice(notice_id)
        return Notice(data)

    async def fetch_member(self, member_id: str) -> Member:
        """Fetches a Weverse Member from Weverse's API using its ID.

        Parameters
        ----------
        member_id: :class:`str`
            The member ID of the member to fetch.

        Returns
        -------
        :class:`.member.Member`
            The :class:`.member.Member` object of the requested Weverse Member.

        Raises
        ------
        NotFound
            If the requested Weverse Member does not exist.
        """
        data = await self.__fetcher.fetch_member(member_id)
        return Member(data)

    async def fetch_comment(self, comment_id: str) -> Comment:
        """Fetches a Weverse Comment from Weverse's API using its ID.

        Parameters
        ----------
        comment_id: :class:`str`
            The comment ID of the comment to fetch.

        Returns
        -------
        :class:`.comment.Comment`
            The :class:`.comment.Comment` object of the requested Weverse Comment.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the comment belongs to.
        NotFound
            If the requested Weverse Comment does not exist.
        """
        data = await self.__fetcher.fetch_comment(comment_id)
        return Comment(data)

    async def fetch_artist_comments(self, post_id: str) -> list[Comment]:
        """Fetches a list of Weverse Artist Comments from a specific post from
        Weverse's API using the post ID.

        Parameters
        ----------
        post_id: :class:`str`
            The post ID of the post to fetch the artist comments from.

        Returns
        -------
        list[:class:`.comment.Comment`]
            The list that contains the :class:`.comment.Comment` objects of the
            Weverse Artist Comment that belong to the specified post.

        Raises
        ------
        Forbidden
            If the signed in account has not joined the requested Weverse Community
            the post belongs to.
        NotFound
            If the requested Weverse Post does not exist.
        """
        data = await self.__fetcher.fetch_artist_comments(post_id)
        data = data["data"]
        artist_comments = [Comment(comment_data) for comment_data in data]
        return artist_comments

    async def start(self) -> asyncio.Task:
        """Starts the Weverse Client."""
        await self.__fetcher.fetch_access_token()
        await self.fetch_new_notifications()  # Populates the cache.
        logger.info(
            "Started listening for Weverse Notifications. "
            "You will be updated every %s seconds.",
            self.stream_interval,
        )
        self.__task = asyncio.create_task(self.__notification_stream())
        return self.__task

    def stop(self) -> None:
        """Disconnects the Weverse Client."""
        self.__task.cancel()
        self.__task = None
        logger.info("The Weverse Client has been stopped.")

    async def __notification_stream(self) -> None:
        while True:
            try:
                await asyncio.sleep(self.stream_interval)
                notifications, comments = await self.fetch_new_notifications()

                for notification in notifications:
                    await self.on_new_notification(notification)

                    if notification.post_type == NotificationType.POST:
                        post = await self.fetch_post(notification.post_id)
                        await self.on_new_post(post)

                    elif notification.post_type == NotificationType.MOMENT:
                        moment = await self.fetch_moment(notification.post_id)
                        await self.on_new_moment(moment)

                    elif notification.post_type == NotificationType.MEDIA:
                        media = await self.fetch_media(notification.post_id)
                        if media:  # Don't call this method when the media is paid.
                            await self.on_new_media(media)

                    elif notification.post_type == NotificationType.LIVE:
                        live = await self.fetch_live(notification.post_id)
                        await self.on_new_live(live)

                    elif notification.post_type == NotificationType.NOTICE:
                        notice = await self.fetch_notice(notification.post_id)
                        if notice:  # Don't call this method for non-artist notice.
                            await self.on_new_notice(notice)

                for comment in comments:
                    await self.on_new_comment(comment)

            except Exception as exception:
                await self.on_exception(exception)

    async def on_new_notification(self, notification: Notification) -> None:
        """This method is called when a new notification is detected.
        You can overwrite this method to do what you want with the new
        notification.
        """

    async def on_new_post(self, post: Post) -> None:
        """This method is called when a new post is detected.
        You can overwrite this method to do what you want with the new post.
        """

    async def on_new_moment(self, moment: Moment | OldMoment) -> None:
        """This method is called when a new moment is detected.
        You can overwrite this method to do what you want with the new moment.
        """

    async def on_new_media(
        self, media: ImageMedia | WeverseMedia | YoutubeMedia
    ) -> None:
        """This method is called when a new media is detected.
        You can overwrite this method to do what you want with the new media.
        """

    async def on_new_live(self, live: Live) -> None:
        """This method is called when a new live is detected.
        You can overwrite this method to do what you want with the new live.
        """

    async def on_new_notice(self, notice: Notice) -> None:
        """This method is called when a new notice is detected.
        You can overwrite this method to do what you want with the new notice.
        """

    async def on_new_comment(self, comment: Comment) -> None:
        """This method is called when a new comment is detected.
        You can overwrite this method to do what you want with the new comment.
        """

    async def on_exception(self, exception: Exception) -> None:
        """This method is called when an unhandled exception occurs.
        You can overwrite this method to do what you want with the exception.
        By default, an exception message is logged.
        """
        logger.exception(exception)


class _WeverseCache:
    def __init__(self):
        self.cache_limit: int = 5000
        self.notification_cache: dict[int, Notification] = {}
        self.comment_notification_cache: dict[str, int] = {}
        self.comment_cache: dict[str, Comment] = {}

    def cache_comment_noti(self, comment_notification: Notification) -> None:
        """Caches comment notifications so that they can be used
        to determine how many and which comment notifications are new.

        Parameters
        ----------
        comment_notification: :class:`Notification`
            The comment notification to cache.
        """
        if len(self.comment_notification_cache) >= self.cache_limit:
            del self.comment_notification_cache[
                list(self.comment_notification_cache.keys())[0]
            ]

        self.comment_notification_cache[
            comment_notification.post_id + comment_notification.author.id
        ] = comment_notification.count

    def cache_notification(self, notification: Notification) -> None:
        """Caches notifications so that they can be used to determine which
        notifications are new.

        Parameters
        ----------
        notification: :class:`Notification`
            The notification to cache.
        """
        if len(self.notification_cache) >= self.cache_limit:
            del self.notification_cache[list(self.notification_cache.keys())[0]]

        self.notification_cache[notification.id] = notification

    def cache_comment(self, comment: Comment) -> None:
        """Caches comments so that they can be used to determine which
        comments are new.

        Parameters
        ----------
        comment: :class:`Comment`
            The comment to cache.
        """
        if len(self.comment_cache) >= self.cache_limit:
            del self.comment_cache[list(self.comment_cache.keys())[0]]

        self.comment_cache[comment.id] = comment



from .errors import *
from .fetcher import *
from .url import *
from .utils import *
from .weverse import *

__title__ = 'Weverse.py'
__author__ = 'Anson Quek'
__license__ = 'MIT'
__version__ = '1.0.0.0'



class PartialMember:
    """Represents a Weverse Member with partial information available.

    Inherited by:

    - :class:`Artist`
    - :class:`PostAuthor`
    - :class:`Member`

    .. container:: operations

        .. describe:: x == y

            Checks if two members are equal.

        .. describe:: x != y

            Checks if two members are not equal.

        .. describe:: hash(x)

            Returns the member's hash.

        .. describe:: str(x)

            Returns the member's name.

    Attributes
    ----------
    data: :class:`str`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`str`
        The ID of the member.
    name: :class:`str`
        The name of the member.
    image_url: :class:`str` | :class:`None`
        The URL of the profile image of the member, if any.
    profile_type: :class:`str`
        The profile type of the member.
    """

    __slots__ = ("data", "id", "name", "image_url", "profile_type")

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: str = data["memberId"]
        self.name: str = data["profileName"]
        self.image_url: str | None = data.get("profileImageUrl")
        self.profile_type: str = data["profileType"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Partial Member member_id={self.id}, name={self.name}"

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.id)


class Artist(PartialMember):
    """Represents a Weverse Artist that is fetched from the
    :meth:`weverse.WeverseClient.fetch_artists()` method.
    Inherits from :class:`PartialMember`.

    Shares the same attributes and operations with :class:`PartialMember`.

    Attributes
    ----------
    community_id: :class:`int`
        The community ID of the community the artist belongs to.
    artist_profile: :class:`ArtistProfile`
        The :class:`ArtistProfile` object of the artist.
    join_date: :class:`int`
        The epoch time when the artist joined the community.
    """

    __slots__ = ("community_id", "artist_profile", "join_date")

    def __init__(self, data: dict):
        super().__init__(data)
        self.community_id: int = data["communityId"]
        self.artist_profile: ArtistProfile = ArtistProfile(
            data["artistOfficialProfile"]
        )
        self.join_date: int = data["joinedDate"]

    def __repr__(self):
        return f"Artist member_id={self.id}, name={self.name}"


class ArtistProfile:
    """Represents an Official Weverse Artist Profile.

    .. container:: operations

        .. describe:: str(x)

            Returns the official name of the artist.

    Attributes
    ----------
    official_name: :class:`str`
        The official name of the artist.
    official_image_url: :class:`str`
        The URL of the official image of the artist.
    birthday: :class:`int`
        The epoch time of the artist's birthday.
    """

    __slots__ = ("official_name", "official_image_url", "birthday")

    def __init__(self, data: dict):
        self.official_name: str = data["officialName"]
        self.official_image_url: str = data["officialImageUrl"]
        self.birthday: int = data["birthday"]["date"]

    def __repr__(self):
        return f"Artist Profile official_name={self.official_name}"

    def __str__(self):
        return self.official_name


class PostAuthor(PartialMember):
    """Represents a Weverse Post Author that is an attribute belonging to :class:`.post.Post`
    which is fetched from the :meth:`weverse.WeverseClient.fetch_post()` method.
    Inherits from :class:`PartialMember`.

    Shares the same attributes and operations with :class:`PartialMember`.

    Attributes
    ----------
    community_id: :class:`int`
        The community ID of the community the post author belongs to.
    has_joined: :class:`bool`
        Whether the post author has joined the community.
    is_official: :class:`bool`
        Whether the post author is an official Weverse account.
    profile_is_accessible: :class:`bool`
        Whether the profile of the post author is accessible.
    is_my_profile: :class:`bool`
        Whether the post author is yourself.
    artist_profile: :class:`ArtistProfile` | :class:`None`
        The :class:`ArtistProfile` object of the Post Author if it's an Artist.
    """

    __slots__ = (
        "community_id",
        "has_joined",
        "is_official",
        "profile_is_accessible",
        "is_my_profile",
        "artist_profile",
    )

    def __init__(self, data: dict):
        super().__init__(data)
        self.community_id: int = data["communityId"]
        self.has_joined: bool = data["joined"]
        self.is_official: bool = data["hasOfficialMark"]
        self.profile_is_accessible: bool = data["profileSpaceStatus"] == "ACCESSIBLE"
        self.is_my_profile: bool = data["myProfile"]
        self.artist_profile: ArtistProfile | None = (
            ArtistProfile(data["artistOfficialProfile"])
            if "artistOfficialProfile" in data
            else None
        )

    def __repr__(self):
        return f"Post Author member_id={self.id}, name={self.name}"


class Member(PartialMember):
    """Represents a Weverse Member that is fetched from the
    :meth:`weverse.WeverseClient.fetch_member()` method.
    Inherits from :class:`PartialMember`.

    Shares the same attributes and operations with :class:`PartialMember`.

    Attributes
    ----------
    community_id: :class:`int`
        The community ID of the community the member belongs to.
    profile_cover_image_url: :class:`str` | :class:`None`
        The URL of the profile cover image of the member, if any.
    profile_comment: :class:`str` | :class:`None`
        The profile comment of the member, if any.
    has_joined: :class:`bool`
        Whether the member has joined the community.
    has_membership: :class:`bool`
        Whether the member has a paid membership to the community.
    is_official: :class:`bool`
        Whether the member is an official Weverse account.
    is_hidden: :class:`bool`
        Whether the member is hidden.
    is_blinded: :class:`bool`
        Whether the member is blinded.
    is_followed: :class:`bool`
        Whether the signed-in account is following the member.
    is_my_profile: :class:`bool`
        Whether the member is yourself.
    first_joined_at: :class:`int`
        The time the member first joined the community at, in epoch.
    follow_count: :class:`int` | :class:`None`
        The number of followers the member has, if available.
    artist_profile: :class:`ArtistProfile` | :class:`None`
        The :class:`ArtistProfile` object of the member if it's an Artist.
    """

    __slots__ = (
        "community_id",
        "profile_cover_image_url",
        "profile_comment",
        "has_joined",
        "has_membership",
        "is_official",
        "is_hidden",
        "is_blinded",
        "is_followed",
        "is_my_profile",
        "first_joined_at",
        "follow_count",
        "artist_profile",
    )

    def __init__(self, data: dict):
        super().__init__(data)
        self.community_id: int = data["communityId"]
        self.profile_cover_image_url: str | None = data.get("profileCoverImageUrl")
        self.profile_comment: str | None = data.get("profileComment")
        self.has_joined: bool = data["joined"]
        self.has_membership: bool = data["hasMembership"]
        self.is_official: bool = data["hasOfficialMark"]
        self.is_hidden: bool = data["hidden"]
        self.is_blinded: bool = data["blinded"]
        self.is_followed: bool = data["followed"]
        self.is_my_profile: bool = data["myProfile"]
        self.first_joined_at: int = data["firstJoinAt"]
        self.follow_count: int | None = (
            data["followCount"]["followerCount"] if "followCount" in data else None
        )
        self.artist_profile: ArtistProfile | None = (
            ArtistProfile(data["artistOfficialProfile"])
            if "artistOfficialProfile" in data
            else None
        )

    def __repr__(self):
        return f"Member member_id={self.id}, name={self.name}"



from .member import PartialMember, PostAuthor


class BaseParent:
    """Represents a Weverse Base Parent class.

    Inherited by:

    - :class:`ParentComment`
    - :class:`ParentPost`
    - :class:`ParentMediaPost`
    - :class:`RootPost`
    - :class:`RootMediaPost`

    Attributes
    ----------
    data: :class:`str`
        The raw data directly taken from the response generated by Weverse's API.
    body: :class:`str`
        The body of the base parent object.
    author: :class:`.member.PartialMember`
        The :class:`.member.PartialMember` object of the author who wrote the
        base parent object.
    """

    __slots__ = ("data", "body", "author")

    def __init__(self, data: dict):
        self.data: dict = data
        self.body: str = data["body"]
        self.author: PartialMember = PartialMember(data["author"])


class ParentComment(BaseParent):
    """Represents a Weverse Parent Comment. Inherits from :class:`BaseParent`.

    Shares the same attributes with :class:`BaseParent`.

    .. container:: operations

        .. describe:: x == y

            Checks if two parent comments are equal.

        .. describe:: x != y

            Checks if two parent comments are not equal.

        .. describe:: hash(x)

            Returns the parent comment's hash.

        .. describe:: str(x)

            Returns the parent comment's body.

    Attributes
    ----------
    id: :class:`str`
        The ID of the parent comment. If more information about the parent
        comment is required by the user, the :meth:`.WeverseClient.fetch_comment()`
        method can be used to fetch the actual :class:`Comment` object.
    created_at: :class:`int`
        The time the parent comment got created at, in epoch.
    """

    __slots__ = ("id", "created_at")

    def __init__(self, data: dict):
        super().__init__(data)
        self.id: str = data["commentId"]
        self.created_at: int = data["createdAt"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Parent Comment comment_id={self.id}, body={self.body}"

    def __str__(self):
        return self.body

    def __hash__(self):
        return hash(self.id)


class ParentPost(BaseParent):
    """Represents a Weverse Parent Post. Inherits from :class:`BaseParent`.

    Shares the same attributes with :class:`BaseParent`.

    Inherited by:

    - :class:`ParentMediaPost`

    .. container:: operations

        .. describe:: x == y

            Checks if two parent posts are equal.

        .. describe:: x != y

            Checks if two parent posts are not equal.

        .. describe:: hash(x)

            Returns the parent post's hash.

        .. describe:: str(x)

            Returns the parent post's plain body.

    Attributes
    ----------
    id: :class:`str`
        The ID of the parent post. If more information about the parent
        post is required by the user, the :meth:`.WeverseClient.fetch_post()`
        method can be used to fetch the actual :class:`.post.Post` object.
    plain_body: :class:`str`
        The plain body of the parent post that does not have markdowns
        and unnecessary information.
    post_type: :class:`str`
        The post type of the parent post.
    """

    __slots__ = ("id", "plain_body", "post_type")

    def __init__(self, data: dict):
        super().__init__(data)
        self.id: str = data["postId"]
        self.plain_body: str = data["plainBody"]
        self.post_type: str = data["postType"]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Parent Post post_id={self.id}, body={self.plain_body}"

    def __str__(self):
        return self.plain_body

    def __hash__(self):
        return hash(self.id)


class ParentMediaPost(ParentPost):
    """Represents a Weverse Parent Media Post. Inherits from
    :class:`BaseParent` and :class:`ParentPost`.

    Shares the same attributes with :class:`BaseParent` and :class:`ParentPost`.

    .. container:: operations

        .. describe:: str(x)

            Returns the parent media post's title.

    Attributes
    ----------
    title: :class:`str`
        The title of the parent media post.
    """

    __slots__ = ("title",)

    def __init__(self, data: dict):
        super().__init__(data)
        self.title: str = data["title"]

    def __repr__(self):
        return f"Parent Media Post post_id={self.id}, title={self.title}"

    def __str__(self):
        return self.title


class RootPost(ParentPost):
    """Represents a Weverse Root Post. Inherits from
    :class:`BaseParent` and :class:`ParentPost`.

    Shares the same attributes with :class:`BaseParent` and :class:`ParentPost`.

    Attributes
    ----------
    url: :class:`str`
        The URL that leads to the root post.
    section_type: :class:`str`
        The section the root post falls under.
    author: :class:`.member.PostAuthor`
        The :class:`.member.PostAuthor` object of the author
        who wrote the root post.
    """

    __slots__ = ("url", "section_type")

    def __init__(self, data: dict):
        super().__init__(data)
        self.url: str = data["shareUrl"]
        self.section_type: str = data["sectionType"]
        self.author: PostAuthor = PostAuthor(data["author"])

    def __repr__(self):
        return f"Root Post post_id={self.id}, body={self.plain_body}"


class RootMediaPost(ParentMediaPost):
    """Represents a Weverse Root Media Post. Inherits from
    :class:`BaseParent`, :class:`ParentPost` and :class:`ParentMediaPost`.

    Shares the same attributes with :class:`BaseParent`, :class:`ParentPost`
    and :class:`ParentMediaPost`.

    Attributes
    ----------
    url: :class:`str`
        The URL that leads to the root media post.
    section_type: :class:`str`
        The section the root media post falls under.
    author: :class:`.member.PostAuthor`
        The :class:`.member.PostAuthor` object of the author
        who wrote the root media post.
    """

    __slots__ = ("url", "section_type")

    def __init__(self, data: dict):
        super().__init__(data)
        self.url: str = data["shareUrl"]
        self.section_type: str = data["sectionType"]
        self.author: PostAuthor = PostAuthor(data["author"])

    def __repr__(self):
        return f"Root Media Post post_id={self.id}, title={self.title}"


class Comment:
    """Represents a Weverse Comment.

    .. container:: operations

        .. describe:: x == y

            Checks if two comments are equal.

        .. describe:: x != y

            Checks if two comments are not equal.

        .. describe:: hash(x)

            Returns the comment's hash.

        .. describe:: str(x)

            Returns the comment's body.

    Attributes
    ----------
    data: :class:`str`
        The raw data directly taken from the response generated by Weverse's API.
    id: :class:`str`
        The ID of the comment.
    body: :class:`str`
        The body of the comment.
    comment_count: :class:`int`
        The number of comments for the comment.
    like_count: :class:`int`
        The number of likes for the comment.
    created_at: :class:`int`
        The time the comment got created at, in epoch.
    author: :class:`.member.PostAuthor`
        The :class:`.member.PostAuthor` object of the author who wrote the comment.
    """

    __slots__ = (
        "data",
        "id",
        "body",
        "comment_count",
        "like_count",
        "created_at",
        "author",
    )

    def __init__(self, data: dict):
        self.data: dict = data
        self.id: str = data["commentId"]
        self.body: str = data["body"]
        self.comment_count: int = data["commentCount"]
        self.like_count: int = data["emotionCount"]
        self.created_at: int = data["createdAt"]
        self.author: PostAuthor = PostAuthor(data["author"])

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id

        raise NotImplementedError

    def __repr__(self):
        return f"Comment comment_id={self.id}, body={self.body}"

    def __str__(self):
        return self.body

    def __hash__(self):
        return hash(self.id)

    @property
    def parent(self) -> ParentComment | ParentMediaPost | ParentPost:
        """:class:`ParentComment` | :class:`ParentPost` | :class:`ParentMediaPost`:

        - Returns the :class:`ParentComment` object if the parent is a comment.
        - Returns the :class:`ParentMediaPost` object if the parent is a media.
        - Returns the :class:`ParentPost` object if the parent is a post.
        """
        post_types = ["NORMAL", "MOMENT", "OFFICIAL"]

        if self.data["parent"]["type"] == "COMMENT":
            return ParentComment(self.data["parent"]["data"])

        if self.data["parent"]["type"] == "POST":
            if self.data["parent"]["data"]["postType"] in post_types:
                return ParentPost(self.data["parent"]["data"])

            else:
                return ParentMediaPost(self.data["parent"]["data"])

    @property
    def root(self) -> RootMediaPost | RootPost:
        """:class:`RootMediaPost` | :class:`RootPost`:

        - Returns the :class:`RootMediaPost` object if the root post is a media post.
        - Returns the :class:`RootPost` object if the root post is a normal post.
        """
        post_types = ["NORMAL", "MOMENT", "OFFICIAL"]

        if self.data["root"]["data"]["postType"] in post_types:
            return RootPost(self.data["root"]["data"])

        return RootMediaPost(self.data["root"]["data"])

    @property
    def url(self) -> str:
        """:class:`str`: Returns the URL to the comment."""
        return self.root.url + f"/comment/{self.id}"
